#!/usr/bin/env python

from math import atan2
from sys import stdout
from os import remove
import time
import tempfile
from ephem import *
from masks import *
from numpy.oldnumeric.mlab import *
from basics import *
from numpy.core.numerictypes import *
import numpy
from VTKHelperFunctions import *
from FITS import *
from NEWFIRM import JTRANS as NEWFIRMJTRANS
from FOURSTAR import JTRANS as FOURSTARJTRANS
import SimpleQueue
from sets import Set
from TwistedHub import *

import ZODB, ZODB.FileStorage, transaction
import persistent
from persistent.dict import *
from persistent.list import *
from BTrees.OOBTree import OOBTree
from BTrees.IOBTree import IOBTree

def GetMJD(utyear,utmonth,utday,uthour,utminute,utsecond):
    a = (14-utmonth)/12
    y = utyear + 4800 - a
    m = utmonth + 12*a - 3
    JDN = utday + (153*m+2)/5 + 365*y + y/4 - y/100 + y/400 - 32045
    JD = JDN + (uthour-12.)/24. + utminute/1440. + utsecond/86400.
    MJD = JD - 2400000.5
    return MJD

class irframe(persistent.Persistent):
    def __repr__(self):
      return "%s[%d(%d)]" % (self.newfile,self.exptime,self.info["NOCCOADD"])

    def order(self,no):
       self.norder = no

    def group(self,t,refdir=None):
       # obsoleted! use FindGroups!
       dx,dy= normalcoord(degrad(t.ra) % (2*pi),degrad(t.dec),\
               degrad(self.ra) % (2*pi),degrad(self.dec), conv=180.0*3600.0/pi)
       if t.fclass == "newfirm":
          NSCALE = 0.4
          gtol = 300.0
       elif t.fclass == "fourstar":
          NSCALE = 0.16
          gtol = 120.0
       m0 = abs(dx) < gtol
       m1 = abs(dy) < gtol
       m2 = (self.object == t.object)
       m3 = (self.filter == t.filter)

       if m0 and m1 and m2 and m3:
           if refdir: self.refgroups.append( refdir+"/"+t.newfile)
           else:
              self.groups.append( t.newfile)
              self.groupe.append( t.exptime)
              self.groupx.append( -dx/NSCALE)
              self.groupy.append( dy/NSCALE)

    def clean(self,t):
       if t != self and t.groups == self.groups: t.groups=[]

    def set(self,t):
       m0 = less(abs(self.mjd - t.mjd),0.4)
       m1 = (self.type == t.type)
       m2 = (self.object == t.object)
       m3 = (self.filter == t.filter)
       if m0 and m1 and m2 and m3: self.sets += 1

class fourstar(irframe):
    def __init__(self,record,keys):
       self.fclass = "fourstar"
       if type(record) != ListType:
           simplify = lambda l: l.replace(" ",",")
           record = simplify(record)
           simplify = lambda l: l.replace(",,",",")
           while record.count(",,"):
              record = simplify(record)
           record = record.replace(",',","',")
           record = record[:-3]+"'"
           record = record.replace("'0,[s]'","0.0")
           record = record.replace(",[s]'","'")
           record = record.split(",")
           record[1:] = map(eval, record[1:])
       self.info = {}
       [self.info.setdefault(key,rec) for rec,key in zip(record,keys)]
       utyear,utmonth,utday= map(int,self.info["UT-DATE"].split("-"))
       uttime = self.info["UT-TIME"].split(":")
       uthour,utminute,utsecond= int(uttime[0]),int(uttime[1]),float(uttime[2])
       MJD = GetMJD(utyear,utmonth,utday,uthour,utminute,utsecond)
       self.info["MJD-OBS"] = MJD
       if self.info["FILTER"] == "undef":
          if self.info["WHEEL-T"] == "Open":
              self.info["FILTER"] = self.info["WHEEL-B"]
          elif self.info["WHEEL-B"] == "Open":
              self.info["FILTER"] = self.info["WHEEL-T"]
          elif "NB" in self.info["WHEEL-T"] and self.info["WHEEL-B"] != "J1":
              self.info["FILTER"] = self.info["WHEEL-T"]
          elif "NB" in self.info["WHEEL-T"] and self.info["WHEEL-B"] == "J1":
              self.info["FILTER"] = "dark"
       if type(self.info["FILTER"]) not in [StringType]:
          self.info["FILTER"] = str( self.info["FILTER"])
       if type(self.info["ASECS"]) in [StringType] and self.info["ASECS"] != "N/A":
          self.info["ASECS"] = float(self.info["ASECS"].split()[0])
       if type(self.info["DSECS"]) in [StringType] and self.info["DSECS"] != "N/A":
          self.info["DSECS"] = float(self.info["DSECS"].split()[0])
       self.base = "_".join(os.path.basename(self.info["Filename"]).replace(".fits","").split("_")[:-1])+"_"
       self.name = "_".join(os.path.basename(self.info["Filename"]).replace(".fits","").split("_")[:-1])+"_.fits"
       self.seq = "_".join(os.path.basename(self.info["Filename"]).replace(".fits","").split("_")[:-2])+"_"
       self.lseq = int(os.path.basename(self.info["Filename"]).replace(".fits","").split("_")[-2])
       self.bdir = os.path.dirname(self.info["Filename"])
#       print self.seq, self.lseq
#       print self.name,MJD,utyear,utmonth,utday,uthour,utminute,utsecond
#       print self.name,
#       print self.info["RA"],
#       print self.info["DEC"]
       if self.info["RA"] != "N/A" and self.info["DEC"] != "N/A" and self.info["MJD-OBS"] != "N/A":
          self.ra = degrad(float(self.info["RA"]))
          self.dec = degrad(float(self.info["DEC"]))
          rao = self.info["ASECS"]/cos(self.dec)
          deco = self.info["DSECS"]
          self.mjd = self.info["MJD-OBS"]
          self.when = int(self.mjd)
       else:
          self.ra = 0.0
          self.dec = 0.0
          rao = 0.0
          deco = 0.0
          self.mjd = 0.0
          self.when = 0
       rint = 6.0*60.0
       rint2 = rint/2.
#       ira0 = int((raddeg(self.ra)*3600-rao/cos(self.dec))/(rint2/cos(self.dec))+0.5)
#       idec0 = int((raddeg(self.dec)*3600-deco)/rint2+0.5)
       ira0 = int((raddeg(self.ra)*3600-rao/cos(self.dec))+0.5)
       idec0 = int((raddeg(self.dec)*3600-deco)+0.5)
#       self.ra0 = ira0*int(rint2/cos(self.dec)+0.5)/3600.0/15.0
#       self.dec0 = idec0*rint2/3600.0
       self.ra0 = ira0/3600.0/15.0
       self.dec0 = idec0/3600.0
       self.RA0 = formatSexagesimal(self.ra0,4,3600).strip()
       self.DEC0 = formatSexagesimal(self.dec0,4,3600).strip()
       if self.DEC0[0] != "-": self.DEC0 = "+"+ self.DEC0
       if abs(raddeg(self.dec)) < 10: self.DEC0=self.DEC0[0] + "0"+self.DEC0[1:]
       if self.RA0.find(":") == 1: self.RA0 = "0"+self.RA0
       self.RA0 = self.RA0.replace(":","")
       self.DEC0 = self.DEC0.replace(":","")
       self.type = self.info["OBSTYPE"]
       self.nloops = self.info["NLOOPS"]
       self.objorig = self.info["OBJECT"]
#       print self.name,
#       print self.info["FILTER"]
       space=Circumstance()
       space.timezone=4.0
       tz=space.timezone
       space.mjd=fromGregorian(utmonth,utday,utyear)+tz/24.
       space.latitude=degrad(scanSexagesimal("-29:00:13"))
       space.longitude=-degrad(scanSexagesimal("70:42:7.1"))
       space.elevation=(EarthRadius+2282.0)/EarthRadius
       space.temperature=0.0
       space.pressure=615.0
       space.epoch=EOD
       twilight=computeTwilight(space, degrad(6.0))
       sr=RiseSet()
       sunob=Obj()
       sunob.any.type=PLANET
       sunob.pl.code=SUN
       computeLocation(space,sunob)
       computeRiseSet(space,sunob,degrad(1.0),sr)
       uttime = scanSexagesimal(self.info["UT-TIME"])
       sunset = mod(toHourOfDay(sr.setTime)-tz,24)
       etwi  = mod(toHourOfDay(twilight[1])-tz,24)
       mtwi = mod(toHourOfDay(twilight[0])-tz,24) + 24.0
       sunrise = mod(toHourOfDay(sr.riseTime)-tz,24) + 24.0
#       mtwi = sunrise-0.5
#       etwi = sunset+0.5
       lctime = uttime - tz + 24.0
#       if sunset-1<lctime<etwi or mtwi<lctime<sunrise+1:
       if (sunset<lctime<etwi and "Ks" not in self.info["FILTER"]) or mtwi<lctime<sunrise or self.type=="tflat":
#       if (etwi<lctime<mtwi):
          self.twilight = True
#          self.type = "tflat"
       else:
          self.twilight = False
          if self.type == "tflat": self.type = "object"
#       if self.twilight:
#          print self.name.split(".")[0],self.twilight,
#          print "%9.3f %9.3f %9.3f %9.3f %9.3f"%(lctime,sunset,etwi,mtwi,sunrise)
       if self.type == "dark" and "dark" not in self.info["FILTER"].lower():
          self.type="astro"
       if "dark" in self.info["FILTER"].lower(): self.type="dark";self.info["FILTER"] = "Dark"
       if "focus" in self.info["OBJECT"].lower() or "FOCUS" in self.info["MACROFIL"]: self.type="focus"
#       if "flat" in self.info["OBSTYPE"].lower(): self.type = "flat"
#       if "flat" in self.info["OBJECT"].lower(): self.type = "flat"
       if "sky" in self.info["OBSTYPE"].lower(): self.sky = True
       else: self.sky = False
       if self.type =="tflat": self.pref = "flt"
       elif self.type in ["test","astro","sky"]: self.pref = "obj"; self.type="object"
       elif self.type =="focus": self.pref = "foc"
       elif self.type =="dark": self.pref = "drk"
       else: self.pref = "unk"
       if self.info["EXPTIME"] == 0:
          self.sky=False
          self.type="test"
       if self.type in ["flat","tflat"]:
          if "ON" in self.info["OBJECT"]: self.info["LAMPSTAT"] = "A"
          elif "OFF" in self.info["OBJECT"]: self.info["LAMPSTAT"] = "B"
          else: self.info["LAMPSTAT"] = ""
       if self.pref == "obj": self.object = "%s%s%s_%s" % (self.pref,self.RA0,self.DEC0,self.info["FILTER"])
#       elif self.pref == "flt": self.object = "flat"+self.info["FILTER"]+self.info["LAMPSTAT"]
#       elif self.pref == "flt": self.object = self.info["OBJECT"].lower()+self.info["LAMPSTAT"]
       elif self.pref == "foc": self.object = "focus"
       elif self.pref == "drk": self.object = "dark%sx%s" % (self.info["EXPTIME"],self.nloops)
#       elif self.pref == "drk": self.object = "dark"
       elif self.pref == "unk": self.object = "unknown"
       self.obsday = int(MJD-tz/12.0+0.5)
#       print self.name,self.object
#       if "flat" in self.info["OBJECT"] or self.type in ["flat","tflat"]:
       if self.type in ["tflat"]:
          self.object = "flat_obj%s" %(self.info["FILTER"])
       if self.object not in ["focus","unknown"]:
           self.object = self.object + "_%s%d" % (self.info["GAIN"][0],self.obsday)
       self.filter = self.info["FILTER"]
       self.jno = int(self.base.split("_")[1])
       self.kno = int(self.base.split("_")[2])
       self.exptime = self.info["EXPTIME"]
       self.gainmode = self.info["GAIN"]
       if self.exptime == "N/A":
          self.exptime = 0
       self.asecs = self.info["ASECS"]
       self.dsecs = self.info["DSECS"]
       self.info["RAOFF"] = self.asecs
       self.info["DECOFF"] = self.dsecs
       self.info["NOCCOADD"] = 1
       self.newfile = self.base
       self.skys=[]
       self.groupt=[]
       self.groupm=[]
       self.groups=[]
       self.groupe=[]
       self.groupk=[]
       self.groupx=[]
       self.groupy=[]
       self.refgroups=[]
       self.sets=0
       mode1s = glob("%s/mode1/%sc?.fits.bad" % (self.object, self.name.replace(".fits","")))
       if mode1s: self.bad = True
       else: self.bad = False
       self.fclass = "fourstar"


class newfirm(irframe):
    def __init__(self,record,keys):
       self.fclass = "newfirm"
       if type(record) != ListType:
           simplify = lambda l: l.replace(" ",",")
           record = simplify(record)
           simplify = lambda l: l.replace(",,",",")
           while record.count(",,"):
              record = simplify(record)
           record = record.replace(",',","',")
           record = record[:-3]+"'"
           record = record.replace("'0,[s]'","0.0")
           record = record.replace(",[s]'","'")
           record = record.split(",")
           record[1:] = map(eval, record[1:])
       self.info = {}
       [self.info.setdefault(key,rec) for rec,key in zip(record,keys)]
       utyear,utmonth,utday= map(int,self.info["DATE-OBS"].split("T")[0].split("-"))
       uttime = self.info["DATE-OBS"].split("T")[1].split(":")
       uthour,utminute,utsecond= int(uttime[0]),int(uttime[1]),float(uttime[2])
       MJD = GetMJD(utyear,utmonth,utday,uthour,utminute,utsecond)
       if self.info["NOCCOADD"] == "N/A":
          self.info["NOCCOADD"] = 1
       if type(self.info["FILTER"]) not in [StringType]:
          self.info["FILTER"] = str( self.info["FILTER"])
       if type(self.info["RAOFF"]) in [StringType] and self.info["RAOFF"] != "N/A":
          self.info["RAOFF"] = float(self.info["RAOFF"].split()[0])
       if type(self.info["RAOFF"]) in [StringType] and self.info["RAOFF"] == "N/A":
          if not isinstance(self.info["NOCDDOF"],StringType):
             self.info["RAOFF"] = float(self.info["NOCDROF"])
       if type(self.info["DECOFF"]) in [StringType] and self.info["DECOFF"] != "N/A":
          self.info["DECOFF"] = float(self.info["DECOFF"].split()[0])
       if type(self.info["DECOFF"]) in [StringType] and self.info["DECOFF"] == "N/A":
          if not isinstance(self.info["NOCDDOF"],StringType):
             self.info["DECOFF"] = float(self.info["NOCDDOF"])
       self.base = os.path.basename(self.info["Filename"]).replace(".fits","")
       self.name = os.path.basename(self.info["Filename"])
#       if "image117992" in self.name:
#           print self.info
#           print MJD
       self.seq = self.base
#       self.lseq = self.base
       self.lseq = 1
       self.bdir = os.path.dirname(self.info["Filename"])
#       print self.name,
#       print self.info["RA"],
#       print self.info["DEC"]
       if self.info["RA"] not in  ["N/A","NC"] and \
                self.info["DEC"] not in ["N/A","NC"] and \
                self.info["RAOFF"] not in ["N/A","NC"] and \
                self.info["DECOFF"] not in ["N/A","NC"] and \
                self.info["MJD-OBS"] != "N/A":
          self.ra = hrrad(scanSexagesimal(self.info["RA"]))
          self.dec = degrad(scanSexagesimal(self.info["DEC"]))
          rao = self.info["RAOFF"]/cos(self.dec)
          deco = self.info["DECOFF"]
          self.mjd = self.info["MJD-OBS"]
          self.when = int(self.mjd)
       else:
          self.ra = 0.0
          self.dec = 0.0
          rao = 0.0
          deco = 0.0
          self.mjd = MJD
          self.when = 0
       rint = 10.0*60.0
       ira0 = int((raddeg(self.ra)*3600-rao)/rint+0.5)
       idec0 = int((raddeg(self.dec)*3600-deco)/rint+0.5)
       self.ra0 = ira0*rint/3600.0/15.0
       self.dec0 = idec0*rint/3600.0
       self.RA0 = formatSexagesimal(self.ra0,4,3600).strip()
       self.DEC0 = formatSexagesimal(self.dec0,4,3600).strip()
       if self.DEC0[0] != "-": self.DEC0 = "+"+ self.DEC0
       if abs(raddeg(self.dec)) < 10: self.DEC0=self.DEC0[0] + "0"+self.DEC0[1:]
       if self.RA0.find(":") == 1: self.RA0 = "0"+self.RA0
       self.RA0 = self.RA0.replace(":","")
       self.DEC0 = self.DEC0.replace(":","")
       self.type = self.info["OBSTYPE"]
       self.objorig = self.info["OBJECT"]
       space=Circumstance()
       space.timezone=4.0
       tz=space.timezone
       space.mjd=fromGregorian(utmonth,utday,utyear)+tz/24.
       space.latitude=degrad(scanSexagesimal("-29:00:13"))
       space.longitude=-degrad(scanSexagesimal("70:42:7.1"))
       space.elevation=(EarthRadius+2282.0)/EarthRadius
       space.temperature=0.0
       space.pressure=615.0
       space.epoch=EOD
       twilight=computeTwilight(space, degrad(6.0))
       sr=RiseSet()
       sunob=Obj()
       sunob.any.type=PLANET
       sunob.pl.code=SUN
       computeLocation(space,sunob)
       computeRiseSet(space,sunob,degrad(1.0),sr)
#       uttime = scanSexagesimal(self.info["UT-TIME"])
       uttime = uthour+utminute/60.0+utsecond/3600.0
       sunset = mod(toHourOfDay(sr.setTime)-tz,24)
       etwi  = mod(toHourOfDay(twilight[1])-tz,24)
       mtwi = mod(toHourOfDay(twilight[0])-tz,24) + 24.0
       sunrise = mod(toHourOfDay(sr.riseTime)-tz,24) + 24.0
#       mtwi = sunrise-0.5
#       etwi = sunset+0.5
       lctime = uttime - tz + 24.0
#       if sunset-1<lctime<etwi or mtwi<lctime<sunrise+1:
       if (sunset<lctime<etwi and "Ks" not in self.info["FILTER"]) or mtwi<lctime<sunrise:
#       if (etwi<lctime<mtwi):
          self.twilight = False
#          self.type = "tflat"
       else:
          self.twilight = False
          if self.type == "tflat": self.type = "object"
#       print self.name,
#       print self.info["FILTER"]
       if "Dark" in self.info["FILTER"] or "dark" in self.info["OBJECT"].lower(): self.type="dark"
       if "FOCUS" in self.info["NOCDHS"]: self.type="focus"
       if "flat" in self.info["OBJECT"].lower(): self.type = "flat"
       if "sky" in self.info["OBSTYPE"].lower(): self.sky = True
       else: self.sky = False
       if self.type =="flat": self.pref = "flt"
       elif self.type =="object": self.pref = "obj"
       elif self.type =="focus": self.pref = "foc"
       elif self.type =="dark": self.pref = "drk"
       else: self.pref = "unk"
       if self.pref == "obj": self.object = "%s%s%s_%s" % (self.pref,self.RA0,self.DEC0,self.info["FILTER"])
       elif self.pref == "flt": self.object = "flat"+self.info["FILTER"]+self.info["LAMPSTAT"]
#       elif self.pref == "flt": self.object = self.info["OBJECT"].lower()+self.info["LAMPSTAT"]
       elif self.pref == "foc": self.object = "focus"
       elif self.pref == "drk":
          self.object = "dark%sx%s" % (self.info["NOCTIM"],self.info["NOCCOADD"])
       elif self.pref == "unk": self.object = "unknown"
       self.obsday = int(MJD-tz/12.0+0.5)
       if self.object not in ["focus","unknown"]:
           self.object = self.object + "_%d" % (self.obsday)
       self.filter = self.info["FILTER"]
#       if "image117992" in self.name:
#           print self.type
#           print self.object
#       self.jno = int(self.base[-4:])
       self.jno = int("".join([l for l in self.base if l not in letters+"./,"]))
       self.kno = 1
       self.exptime = self.info["EXPTIME"]
       if self.exptime == "N/A":
          self.exptime = 0
       self.asecs = self.info["RAOFF"]
       self.dsecs = self.info["DECOFF"]
       self.newfile = self.base
       self.skys=[]
       self.groupt=[]
       self.groupm=[]
       self.groups=[]
       self.groupe=[]
       self.groupk=[]
       self.groupx=[]
       self.groupy=[]
       self.refgroups=[]
       self.sets=0
       mode1s = glob("%s/mode1/%sc?.fits.bad" % (self.object, self.name.replace(".fits","")))
       if mode1s: self.bad = True
       else: self.bad = False
       self.fclass = "newfirm"

def readdb(db,targ,kind,fclass=newfirm):
   f=open(db,"r")
   df = f.readlines()
   f.close()
   header = df[1][1:].replace(" ","").replace("''","','")
   keys = eval(header)
   j = nonzero([(l[0] not in ["#","["]) for l in df])
   df = [df[k] for k in j]
   t = [fclass(df[j],keys) for j in range(len(df))]
   t = [i for i in t if ((targ and targ in i.object) or not targ) and (kind and kind==i.pref or not kind)]
   if targ: [setattr(i,"object",targ) for i in t]
   return t

def reduce2groups(t):
   [[i.clean(x) for x in t] for i in t]
   u = nonzero([i.groups != [] for i in t])
   t = [t[j] for j in u]
   return t

NEWFIRMDBKEYS = ["DATE-OBS","MJD-OBS","EXPTIME","RA","DEC","FILTER","RAOFF","DECOFF","OBSTYPE","OBJECT",
        "LAMPSTAT","NOCDHS","NOCTIM","NOCCOADD","NOCDROF","NOCDDOF"]
FOURSTARDBKEYS = ["UT-DATE","UT-TIME","EXPTIME","RA","DEC","FILTER","ASECS","DSECS","OBSTYPE","OBJECT",
        "LAMPSTAT","GAIN","EGAIN","ENOISE","LOOP","NLOOPS","MACROFIL","WHEEL-T","WHEEL-B","WHEEL-F"]

@defer.inlineCallbacks
def FindGroups(t,G,useint=0):
   n = len(t)
   ra = array([i.ra for i in t])
   dec = array([i.dec for i in t])
   et = array([i.exptime for i in t])
   fno = array([int(i.jno) for i in t])
   gno = array([int(i.kno) for i in t])
   tfilters = array([hash(t1.filter) for t1 in t])
   ttypes = array([hash(t1.type) for t1 in t])
   tskys = array([hash(t1.sky) for t1 in t])
   tobjorigs = array([hash(t1.objorig) for t1 in t])
   if useint:
      tintervals = array([hash(t1.interval) for t1 in t])
   x, y = ra, dec
   dx = (x[::,NewAxis] - x[NewAxis,::])*cos(y)*180/pi*60.0
   dy = (y[::,NewAxis] - y[NewAxis,::])*180/pi*60.0
   science = equal(ttypes,hash("object"))
   science2 = science[::,NewAxis]*science[NewAxis,::]
   dx = where(science2,dx,0)
   dy = where(science2,dy,0)
   if t[0].fclass == "newfirm":
      tol = 10.0
   elif t[0].fclass == "fourstar":
      tol = 4.0
      tol = 4.0
      sm0 = less(abs(dx),30.0)
      sm1 = less(abs(dy),30.0)
      keepsky = sm0*sm1
      keepsky = keepsky*equal(tobjorigs[::,NewAxis],tobjorigs[NewAxis,::])
      keepsky = keepsky*equal(tfilters[::,NewAxis],tfilters[NewAxis,::])
      keepsky = keepsky*logical_or(tskys[::,NewAxis],tskys[NewAxis,::])
#      keepsky = keepsky*array([[((i1.sky or i2.sky) and (i1.filter == i2.filter) and (i1.objorig == i2.objorig)) for i2 in t] for i1 in t])
      tol = where(keepsky,120.0,tol)
   m0 = less(abs(dx),tol)
   m1 = less(abs(dy),tol)
   m2 = equal(tfilters[::,NewAxis],tfilters[NewAxis,::])
#   m2 = array([[(t[i].filter==t[j].filter) for j in range(n)] for i in range(n)])
   m3 = equal(ttypes[::,NewAxis],ttypes[NewAxis,::])
#   m3 = array([[(t[i].type==t[j].type) for j in range(n)] for i in range(n)])
   m3b = equal(ttypes[::,NewAxis],ttypes[NewAxis,::])*equal(ttypes,hash("tflat"))[::,NewAxis]
#   m3b = array([[(t[i].type==t[j].type and t[i].type=="tflat") for j in range(n)] for i in range(n)])
   m3c = equal(ttypes[::,NewAxis],ttypes[NewAxis,::])*not_equal(ttypes,hash("tflat"))[::,NewAxis] * \
        less(abs(getlog10(et[::,NewAxis])-getlog10(et[NewAxis,::])),log10(2))
#   m3c = array([[(t[i].type==t[j].type and t[i].type!="tflat" and abs(getlog10(et[i])-getlog10(et[j]))<log10(2)) for j in range(n)] for i in range(n)])
   U = logical_or(m0*m1*m2*m3*m3c,m2*m3b)
   if useint:
      m4 = equal(tintervals[::,NewAxis],tintervals[NewAxis,::])
#      m4 = array([[(t[i].interval==t[j].interval) for j in range(n)] for i in range(n)])
      U = logical_or(m4*m3*m3c,m3b) * U
   t[0].groupt = [t[0]]
   t[0].groupm = [t[0].mjd]
   t[0].groups = [t[0]]
   t[0].groupe = [t[0].exptime]
   t[0].groupk = [t[0].sky]
   t[0].groupx = [0.0]
   t[0].groupy = [0.0]
   G.append(t[0])
   I = [0]
   see1 = time.time()
   print "Begin Grouping!"
   @defer.inlineCallbacks
   def SeeWhatWeCanDoToo():
       def n2():
           for j in range(n):
              new = 1
              if j in I:
                 new = 0
              else:
                 for i,g in zip(I,G):
                    if t[j] in g.groups:
                      new = 0
                    else:
                      if U[j][i]:
                          new = 0
                          g.groupt += [t[j]]
                          g.groupm += [t[j].mjd]
                          g.groups += [t[j]]
                          g.groupe += [t[j].exptime]
                          g.groupk += [t[j].sky]
                          g.groupx += [-dx[i][j]]
                          g.groupy += [dy[i][j]]
                          yield
              if new:
                 left = 1
#                 print j,t[j].object,t[j].ra,t[j].dec,t[j].RA0,t[j].DEC0
                 t[j].groupt = [t[j]]
                 t[j].groupm = [t[j].mjd]
                 t[j].groups = [t[j]]
                 t[j].groupe = [t[j].exptime]
                 t[j].groupk = [t[j].sky]
                 t[j].groupx = [0.0]
                 t[j].groupy = [0.0]
                 G.append(t[j])
                 I.append(j)
              yield
       co = task.cooperate(n2())
       yield co.whenDone()
   print "SeeWhatWeCanDoToo!"
   co = SeeWhatWeCanDoToo()
   yield co
   see2 = time.time()
   print "SeeWhatWeCanDidToo?", see2-see1

def reduce2groups(t):
   [[i.clean(x) for x in t] for i in t]
   u = nonzero([i.groups != [] for i in t])
   t = [t[j] for j in u]
   return t

@defer.inlineCallbacks
def CollateDB(dbroot,targ="",filt="",nume=0,kind=None,watching=False,fclass=newfirm,ssub=5,read_only=True,plist=[]):
  intarg = targ

  if type(dbroot) is StringType:
     groups = {}
  else:
     print "In CollateDB:",list(dbroot.keys())
     db = dbroot["Data"]
     if "Groups" in dbroot:
        dbgroups = dbroot["Groups"]
        groups = {}
        for key in dbgroups:
           groups[key] = copy.deepcopy(dbgroups[key])
#        print list(groups.keys())
        if intarg:
           print intarg,intarg in list(groups.keys())
     else:
        groups = {}

  if not intarg or intarg not in groups:
     if intarg:
        print "Need to figure out what %s is!" % (intarg)
     targ = targ.replace("sky","obj")
     if "_tmp" in targ and "obj" in targ:
        targ = targ.split("_tmp")[0]

     if type(db) is StringType:
       print "Reading database",db
       ti = readdb(db,targ,kind,fclass=fclass)
     else:
       ti = [db[i] for i in db]
     ti = [copy.deepcopy(t) for t in ti]

     mjdend = ti[-1].mjd
     gmtime = time.gmtime()
     utyear,utmonth,utday= gmtime[0],gmtime[1],gmtime[2]
     uthour,utminute,utsecond=  gmtime[3],gmtime[4],gmtime[5]
     mjdend = GetMJD(utyear,utmonth,utday,uthour,utminute,utsecond)

     tw = []

     see1 = time.time()

     @defer.inlineCallbacks
     def SeeWhatWeCanDo():
         print "In SeeWhatWeCanDo!"
         def iti():
             for i in ti:
                 if i.twilight and i.filter not in ["Dark"]:
                     ni = copy.deepcopy(i)
                     ni.type = "tflat"
                     ni.pref = "flat"
                     ni.object = "flat_obj%s_%s%d" % (ni.filter,ni.gainmode[0],ni.obsday)
                     incl = True
                     for i2 in ti:
                        if repr(ni) == repr(i2) and ni.object == i2.object:
                           incl = False
                           break
#                     print ni,ni.object,incl
                     if incl: tw.append(ni)
                 yield
         co = task.cooperate(iti())
         yield co.whenDone()

     print "SeeWhatWeCanDo!"
     co = SeeWhatWeCanDo()
     yield co
     see2 = time.time()
     print "SeeWhatWeCanDid?",see2-see1

#     print tw

     ti = ti + tw
#     targets = list(Set([i.object for i in ti if i.object]))
#     targets.sort()
#     for j,tttt in enumerate(targets): print j+1,tttt

     t = copy.deepcopy(ti)

     if len(t):
       mjd0 = int([i.mjd for i in t if i.mjd][0])
       mjd1 = 24*(t[-1].mjd-mjd0)
     else:
       [plist.pop(0) for k in range(len(plist))]
       return
#       return {}

     see1 = time.time()
     print "FindGroups!",len(t)
     G = []
     yield FindGroups(t,G)
     see2 = time.time()
     print "FoundGroups!",see2-see1
#     d = defer.Deferred()
#     d.addCallback(lambda result: FindGroups(t,G))
#     bd = threads.deferToThread(d.callback,True)
#     yield d
     t = G

     targets = list(Set([i.object for i in t if i.object]))
     targets.sort()
#     for j,tttt in enumerate(targets): print j+1,tttt
#     print
#     for k in t:
#         print k.object,k.type,k.exptime,len(k.groupm),extrema(k.groupm)

#     os._exit(1)

#     print [t1 for t1 in targets if "obj143840" in t1]

     intervals = []
     tintervals = {}

     see1 = time.time()
     @defer.inlineCallbacks
     def SeeWhatWeCanDo2():
         print "In SeeWhatWeCanDo2!"
         def iti():
             for k in t:
               if k.pref == "obj" and "flat" not in k.object and len(k.groups)>1:
                  dk = array([(k.groupk[j]==k.groupk[j+1]) for j in range(len(k.groupm)-1)])
                  dt = diff(k.groupm)*24*60 - array(k.groupe[:-1])/60.0
                  dt = where(dk,dt,0.0)
                  kdt = compress(dt,dt)
                  bdt,sdt = median(kdt),1.49*median(abs(kdt))
                  sdt = max([sdt,median(array(k.groupe)/60.0)])
                  mdt = max([bdt+10*sdt,10*median(dt)])
                  g = nonzero(greater(dt,mdt))
#                  print dt,len(dt)+1
#                  print g,len(g)
                  if len(dt) not in g: g = concatenate([g,[len(k.groupm)]])
                  ng = len(g)
#                  print g,len(g)
                  for j in range(ng):
                     if j: mjd = k.groupm[g[j-1]+1:g[j]+1]
                     else: mjd = k.groupm[:g[j]+1]
                     mjd = 24*(asarray(mjd) - mjd0)
                     emjd = extrema(mjd)
                     intervals.append(emjd)
                     if k.object not in tintervals: tintervals[k.object] = []
                     tintervals[k.object].append(emjd)
                  print k.object,bdt,sdt,mdt,len(tintervals[k.object])
               yield
         co = task.cooperate(iti())
         yield co.whenDone()
     print "SeeWhatWeCanDo2!"
     co = SeeWhatWeCanDo2()
     yield co
     see2 = time.time()
     print "SeeWhatWeCanDid2?",see2-see1
#     os._exit(1)

     if tintervals:
        dt = []
        ks = []
        for k in tintervals:
           ks.extend(len(tintervals[k])*[k])
           dt.extend(tintervals[k])
        dt = transpose(dt)
        o = argsort(dt[0])
        ks = [ks[o1] for o1 in o]
        dt = take(dt,o,1)

        idt0 = dt[0].astype(int32)
        idt1 = (60*(dt[0]-idt0)).astype(int32)
        jdt0 = dt[1].astype(int32)
        jdt1 = (60*(dt[1]-jdt0)).astype(int32)

     t = copy.deepcopy(ti)
     if tintervals:
        for k in range(len(ks)):
           print ks[k],dt[0][k],dt[1][k]
        print

     bad = []
     nints = {}
     for j,i in enumerate(t):
         dti = (24*(i.mjd-mjd0))
         # Mental note:  did I fuck up dt here?
         # Something is going wrong with the interval adoption during
         # observing!
         gm = ""
         if intervals: dtest = less_equal(dti,dt[1]) * greater_equal(dti,dt[0])
#         if "flat" not in i.object: print i.object,dti,i,
         if i.pref == "obj" and "flat" not in i.object:
            if intervals and i.mjd and dtest.any():
               l = nonzero(dtest)[-1]
               sinterval = "%02d%02d" % (idt0[l],idt1[l])
#               if "obj143840" in i.object:
#               print i.object,dti,i,sinterval,nints.get(sinterval,0),mjdend-i.mjd
               if l == (len(intervals)-0) and (mjdend-i.mjd) < \
                (divz(dt[1][l]-dt[0][l],24*nints.get(sinterval,0))+5*i.exptime/(24*3600)):
                  i.interval = ""
               else:
                  i.interval = sinterval
                  i.jdt0 = jdt0[l]
                  i.jdt1 = jdt1[l]
               i.object = "%s_%s" % (i.object,i.interval)
            else:
               i.interval = ""
            if i.interval:
               if i.interval not in nints: nints[i.interval] = 0
               nints[i.interval] += 1
            else:
#               bad.append(i.object)
               bad.append(i)
         else: i.interval = ""
#         print i,i.object,i.interval
#         if "flat" not in i.object: print i.interval

#     print "FindGroups"
#     print dt[0]
#     print dt[1]
#     print dt[1]-dt[0]
#     os._exit(1)
     t = [i for i in t if i not in bad]
#     print t
#     targets = list(Set([i.object for i in t if i.object]))
#     targets.sort()
#     for j,tttt in enumerate(targets): print j+1,tttt
#     os._exit(1)

     see1 = time.time()
     print "FindGroups!"
     G = []
     yield FindGroups(t,G,useint=1)
     see2 = time.time()
     print "FoundGroups!",see2-see1
     print G
##     d = defer.Deferred()
#     d.addCallback(lambda result: FindGroups(t,G,useint=1))
#     bd = threads.deferToThread(d.callback,True)
#     yield d

#     targets = list(Set([i.object for i in t if i.object]))
#     targets.sort()
#     for j,tttt in enumerate(targets): print j+1,tttt
#     os._exit(1)

#     targets = list(Set([i.object for i in t if i.object not in bad]))
#     G  = [g for g in G if g.object not in bad]
     gtargets = [g.object for g in G]
     gnumbers = [len(g.groupt) for g in G]
     gfilters = [g.filter for g in G]
#     print gtargets
#     print gnumbers
#     print [t for t in targets if "obj143840" in t]
#     os._exit(1)

     groups = {}
     fnume = lambda target: ("flat" in target and 3 or ("dark" in target and 1 or nume))
#     fnume = lambda target: 1
     for group in G:
#         u = nonzero([((targ and targ in i.object) or (target in i.object)) for i in t])
#         u = nonzero([(i.filter == filter1) for i in t])
#         tu = [t[j] for j in u]
         target = group.object
         if (targ and targ in target or not targ) and target not in groups:
            tu = group.groupt
            if filt: tu = [i for i in tu if i.filter == filt]
            jtu = Set([i.jno for i in tu])
            stu = Set([i.jno for i in tu if i.sky])
            if (targ and targ in target or not targ) and len(jtu)>=fnume(target) and len(stu) < len(jtu):
                print "Target %50s [%5d(%5d)]" % (target,len(tu),len(jtu)),tu[0].jno,tu[-1].jno
            [[i.set(x) for x in tu] for i in tu]
            [tu[j].order(j) for j in range(len(tu))]
            if len(jtu)>=fnume(target) and (targ and targ in target or not targ):
               if targ: groups[targ] = tu
               else: groups[target] = tu

     if type(db) is not StringType:
       transaction.abort()

##     os._exit(1)

#     targets = list(Set([i.object for i in t if i.object]))
#     targets.sort()
#     for j,tttt in enumerate(targets): print j+1,tttt
#     os._exit(1)

     sgroups = {}
     for target in groups:
        if "obj" in target:
           t = groups[target]
           t = [copy.deepcopy(i) for i in t if i.sky]
           if len(t):
              sgroups[target.replace("obj","sky")] = t
              for i in t:
                  i.target = target.replace("obj","sky")

     for target in sgroups:
         tu = groups[target] = sgroups[target]
         jtu = Set([i.jno for i in tu])
         print "Target %50s [%5d(%5d)]" % (target,len(tu),len(jtu))

     for target in groups:
         t = groups[target]
         for i in t:
             i.target = target
             if hasattr(i,"subgroups"): delattr(i,"subgroups")

     gtmp = {}
     for target in groups:
         t = groups[target]
         if target == "obj175035-291200_Ks_F56075_2727": verb = 0
         else: verb = 0
         if target[:3] in ["obj"]:
            n = len(t)
            jno = array([i.jno for i in t])
            kno = array([i.kno for i in t])
            s = nonzero(equal(kno,1))
            if verb:
               print jno
               print kno
               print s
            if len(s) > 2*ssub:
               ginc = []
               binc = []
               for b in range(n):
                   bjno = Set([jno[j] for j in binc])
                   if len(bjno) < ssub or kno[b] != 1:
                      binc.append(b)
                   else:
                      ginc.append(binc)
                      binc = [b]
               if binc: ginc.append(binc)
#               print target
               lb = [len(Set([jno[j] for j in binc])) for binc in ginc]
#               print lb
               lb0 = [len([jno[j] for j in binc]) for binc in ginc]
#               print lb
               if lb[-1] < ssub and len(lb) > 1:
                  ginc[-2].extend( ginc.pop(-1))
               if verb:
                  for binc in ginc:
                     print len(binc),binc
#               print lb
               lb = [len(Set([jno[j] for j in binc])) for binc in ginc]
#               print lb
               for binc in ginc:
                   b0 = binc[0]
                   b1 = binc[-1]
                   jn = jno[b0]
                   kn = kno[b0]
                   tt = t[b0:b1+1]
                   if verb:
                       print tt
#                   if hasattr(tt[0],"interval"):
#                      interval = tt[0].interval
#                      initial = interval.split("_")[0]
#                      ttarget = target.replace(interval,initial) + "_tmp%04d" % (jn)
#                   else:
#                      ttarget = target + "_tmp%04d" % (jn)
                   ttarget = target + "_tmp%04d" % (jn)
                   ttmp = []
                   for i in tt:
                       t2 = copy.deepcopy(i)
                       t2.target = ttarget
                       t2.temporary = True
                       ttmp.append(t2)
                   jtt = Set([i.jno for i in ttmp])
                   gtmp[ttarget] = ttmp
               [setattr(i,"subgroups",len(ginc)) for i in t]
#         elif target[:4] in "flat":
#            print target,hasattr(t[0],"subgroups")
#            [delattr(i,"subgroups") for i in t]

     for target in gtmp:
         tu = groups[target] = gtmp[target]
         jtu = Set([i.jno for i in tu])
         print "Target %50s [%5d(%5d)]" % (target,len(tu),len(jtu))

     for target in groups:
         t = groups[target]
         for i in t:
             i.target = target

     deltargs = []
     for target in groups:
         t = groups[target]
         keep = True
         if len(t)>=1:
            j0 = t[0].mjd
            j1 = t[-1].mjd+t[-1].exptime/(24*3600)
            tds = array([t[j].mjd for j in range(len(t))])[1:]
            tde = array([(t[j].mjd + t[j].exptime/(24*3600)) for j in range(len(t))])[:-1]
            tcut = median(diff(tds))
#            print target,tcut*24*60,(mjdend-j1)*24*60,t[-1].exptime/60.,
            if (mjdend - j1) < (3*tcut+t[-1].exptime/(24*3600)):
               keep = False
#            print keep
         else:
            keep = False
         if not keep:
            deltargs.append(target)

     for target in deltargs:
        groups.pop(target)

     if not read_only:
        if "Groups" not in dbroot:
           dbg = OOBTree()
           dbroot["Groups"] = dbg
           transaction.commit()
        for target in groups:
            if target in dbroot["Groups"]:
               if len(dbroot["Groups"][target]) != len(groups[target]):
                  del dbroot["Groups"][target]
                  transaction.commit()
            if target not in dbroot["Groups"]:
               dbroot["Groups"][target] = groups[target]
               transaction.commit()

#  os._exit(1)

#  targets = groups.keys()
#  targets.sort()
#  for j,tttt in enumerate(targets): print j+1,tttt
#  os._exit(1)


  if intarg and intarg in groups:
#     print intarg
#     print list(groups.keys())
#     group = groups[intarg]
#     print "Original group (%d):" % (len(group))
#     for t in group:
#        print t,t.info["Filename"],t.base
     group = []
     infos = []
     [(group.append(t),infos.append(t.base)) for t in groups[intarg] if t.base not in infos]
#     print "Modified group (%d):" % (len(group))
#     for t in group:
#        print t,t.info["Filename"],t.base
#     raw_input("PAUSE")
     [plist.pop(0) for k in range(len(plist))]
     [plist.append(t) for t in group]
     print "Returning from CollateDB"
     return
#     return group
  elif groups or watching:
     [plist.pop(0) for k in range(len(plist))]
     plist.append({})
     for group in groups:
#         print group
         plist[0][group] = groups[group]
     print "Returning from CollateDB"
     return
#     return groups
  else:
     raise NameError("Target %s not found" % (intarg))

def PackZODB(zodb):
    storage = ZODB.FileStorage.FileStorage(zodb)
    zdb = ZODB.DB(storage)
    zdb.pack()
    zdb.close()
    storage.close()

def UpdateZODB(ddir,zodb,newfs=0,verb=0,fclass="newfirm"):
    bad = zodb+"-BAD"
    badness = {}
    if os.path.exists(bad):
      d = {}
      b = open(bad,"r")
      badness = d.fromkeys(b.read().split("\n"))
      b.close()
    if newfs: [os.remove(o) for o in glob(zodb+"*")]
    print "Opening: ",zodb
    storage = ZODB.FileStorage.FileStorage(zodb)
    zdb = ZODB.DB(storage)
    conn = zdb.open()
    dbroot = conn.root()
    if newfs or "Data" not in dbroot:
       tree = OOBTree()
       dbroot["Data"] = tree
       transaction.commit()
    print "In UpdateZODB: ",list(dbroot.keys())
    tree = dbroot["Data"]
    exists = [tree[i].info["Filename"] for i in tree]
    for i in tree:
       if i in badness:
           del tree[i]

    if fclass == "newfirm":
       fitslist = [f for f in glob(ddir+"/image????.fits") if f not in exists]
       fitslist += [f for f in glob(ddir+"/image?????.fits") if f not in exists]
       fitslist += [f for f in glob(ddir+"/image??????.fits") if f not in exists]
       if not fitslist:
          fitslist = [f for f in glob(ddir+"/raw?????.fits") if f not in exists]
       fitslist = [f for f in fitslist if f.split(".")[-2][-1] in digits]
       fitslist.sort()
       DBKEYS = NEWFIRMDBKEYS
    elif fclass == "fourstar":
       fitslist = [f for f in glob(ddir+"/fsr_????_??_c?.fits") if f not in exists]
       uniquelist = list(Set(["_".join(os.path.basename(f).replace(".fits","").split("_")[:-1]) for f in fitslist]))
       fitslist = []
       for f in uniquelist:
         for c in [1,2,3,4]:
           g = "%s/%s_c%d.fits" % (ddir,f,c)
           if os.path.exists(g):
              fitslist.append(g)
              break
       fitslist.sort()
       DBKEYS = FOURSTARDBKEYS

    t0 = time.time()
    def GoThroughFITSList():
        for f in fitslist:
           fb = os.path.basename(f)
           if fb not in tree and not badness.has_key(os.path.basename(f)) and os.stat(f)[6]:
              nf = GetOneFITS(f,fclass,DBKEYS)
              tree[fb] = nf
              transaction.commit()
              if verb:
                 print "Added %s[%s]" % (fb,nf.object)
              yield

    def CloseIt():
       print "Closing: ",zodb
       transaction.commit()
       conn.close()
#       zdb.pack()
       zdb.close()
       storage.close()
       t1 = time.time()
       print "Got all FITS in the list in %.1fs." % (t1-t0)
#       os.remove(zodb+".lock")
#       print glob(zodb+"*")

    co = task.cooperate(GoThroughFITSList())
    co.whenDone().addCallback(lambda result: CloseIt())
    return co.whenDone()

def GetOneFITS(f,fclass,DBKEYS,verb=0):
    fb = os.path.basename(f)
    if verb: print "What is %s?" % (f)
    F = FITS(f)
    fi0 = [F[k] for k in DBKEYS]
    fi = []
    for i in fi0:
      if isinstance(i,StringType): fi.append(i.strip())
      else: fi.append(i)
    fr = [f]+fi
    F.close()
    nf = eval(fclass)(fr,["Filename"]+DBKEYS)
    return nf

if __name__ == "__main__":
    argv,targ = rdarg(argv,"-o",None,None)
    argv,kind = rdarg(argv,"-t",None,None)
    argv,nume = rdarg(argv,"-l",int,0)
    argv,filt = rdarg(argv,"-f",None,None)
    argv,fclass = rdarg(argv,"-c",None,"newfirm")
    db = argv[1]

    plist = []
    CollateDB(db,targ,filt,nume=nume,kind=kind,fclass=fclass,plist=plist)

