#!/usr/bin/env python

from TwistedHub import *

from twisted.python import log as logging
from twisted.application import service, internet

from nevow               import inevow
from nevow               import appserver
from nevow               import athena
from nevow               import rend
from nevow               import loaders
from nevow               import url
from nevow               import tags as T
from nevow               import flat
from nevow               import util
from nevow               import livepage

from twisted.internet import task
import time

from twisted.python.util import sibpath
from twisted.python.filepath import FilePath

sibling = FilePath(__file__).sibling

class MainPage(rend.Page):
    def __init__ ( self, *args, **kwargs ):
        self.hubc = kwargs["hc"]
        kwargs.pop("hc")
        rend.Page.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()

    def child_live(self, ctx): return LivePage()
    def child_hub(self, ctx): return HubPage(hc=self.hubc)

    def makeDocFactory ( self ):
       return loaders.stan(T.html[T.head[T.title["main"]]])

class LivePage(athena.LivePage):
    def __init__ ( self, *args, **kwargs ):
        athena.LivePage.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.jsModules.mapping[u"EchoThing"] = sibling("HubSite.js").path

    def child_(self, ctx): return LivePage()

    def render_echo(self,ctx,data):
        obj = EchoElement()
        obj.setFragmentParent(self)
        return ctx.tag[obj]

    def makeDocFactory ( self ):
        return loaders.stan(T.html[
                T.head(render=T.directive('liveglue'))[
                    T.title["Live Page"],
                    ],
                T.body[
                     T.div(render=T.directive("echo")),
                     ]
                     ])


class EchoElement(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        athena.LiveElement.__init__ ( self, *args, **kwargs )
        self.docFactory = self.makeDocFactory()
        self.index = 0
        self.length = 10

    def makeDocFactory ( self ):
        return loaders.stan(T.div(render=T.directive('liveElement'))[
                     T.h2["Live Element"],
                     T.div[self.say()],
                     T.table(style="border:thin solid black;")[
                     T.thead[T.tr[T.th(style="border:thin solid black;")["I am counting!"]]],
                     T.tbody[T.tr[T.td[T.div(style="height:180px;width:480px;overflow=visible;padding=0px",name="scrollArea")]]]
                     ]
                     ]
                     )

    def say(self):
        print "Starting clock!"
        self.clock = task.LoopingCall(self.loop)
        self.clock.start(0.5)
        return ""

    def beforeRender(self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn(self, reason):
        reactor.callFromThread(self.clock.stop)

    def loop(self):
        if hasattr(self.page,"callRemote"):
           self.index += 1
           if self.index > self.length:
              name = u"clock%d" % (self.index-self.length)
              self.callRemote('delText', u"scrollArea", name)
           name = u"clock%d" % (self.index)
           message = u"I am line number %d" % (self.index)
           self.callRemote('addText', u"div", u"scrollArea", message, name)

class Options(usage.Options):
  optFlags = [
        ["verb", "v", "verbose"],
        ["quiet", "q", "quiet"],
        ["auth", "a", "require hub authorization"],
        ["fast", "f", "don't wait for servers"],
    ]

  optParameters = [
        ["looptime", "l", 0.01, "looptime"],
        ["nice", "n", 10, "Nicing level"],
        ["user", "u", getpass.getuser(), "username"],
        ["port", "p", 9000, "task listening port for hub"],
        ["host", "h", "localhost", "host of running hub"],
        ["sshport", "P", 9022, "ssh listening port for hub"],
        ["machines", "m", None, "comma-separated list of machines"],
        ["import", "i", [], "comma-separated list of modules to import"],
        ["Import", "I", [], "comma-separated list of modules to 'from import *'"],
    ]

  def postOptions(self):
     self["port"] = int( self["port"])
     self["sshport"] = int( self["sshport"])
     self["nice"] = int( self["nice"])
     self["looptime"] = float( self["looptime"])
     if self["machines"]: self["machines"] = self["machines"].split(",")
     else: self["machines"] = MACHINES

options = Options()
options.parseOptions(sys.argv[1:])

user = options["user"]
port = options["port"]
sshport = options["sshport"]
nicelevel = options["nice"]
verb = options["verb"]
auth = options["auth"]
quiet = options["quiet"]
nowait = options["fast"]
host = options["host"]
modules = options["import"]
fmodules = options["Import"]
stime = options["looptime"]

machines = [(":" not in machine and [machine,"1"] or machine.split(":")) for machine in options["machines"]]
machines = [machine[0] for machine in machines for j in range(int(machine[1]))]
print machines

SUBMIT = "_submit"

class HubPage ( athena.LivePage):
    def __init__ ( self, *args, **kwargs ):
        self.hubc = kwargs["hc"]
        kwargs.pop("hc")

        athena.LivePage.__init__ ( self, *args, **kwargs )
        self.jsModules.mapping[u"EchoThing"] = sibling("HubSite.js").path

        self.hub = getattr(self.hubc,"hub",None)
        self.running = 0
        self.start = 0
        self.stop = 0
        self.pause = 0
        self.length = 20
        self.switches = {"Off":(0,1,0),"On":(1,0,0),"Pause":(0,0,1)}
        self.switches["Unpause"] = self.switches["Pause"]
        self.docFactory = self.makeDocFactory()
        self.jsModules.mapping[u"EchoThing"] = sibling("HubSite.js").path

    def locateChild( self, ctx, segments):
        if segments[0] != SUBMIT:
           return athena.LivePage.locateChild(self,ctx,segments)
        request = inevow.IRequest(ctx)
        switch = ctx.arg("switch")
        self.start,self.stop,self.pause = self.switches[switch]

        if self.hubc.running:
           if self.stop:
             self.stop = 0
             self.hubc.hubstop()
           elif self.pause:
             self.pause = 0
             self.hubc.hubpause()
        elif not self.hubc.running and self.start:
           self.start = 0
           self.hubc.hubstart()
        ref = request.getHeader('referer') or ''
        return url.URL.fromString(ref),()

    def render_elements( self, ctx, data):
        if self.hubc.running:
           request = inevow.IRequest(ctx)
           ref = request.getHeader('referer') or ''
           if not getattr(self,"hub",None) or not hasattr(self.hub,"TASKS") or not len(getattr(self.hub,"TASKS")):
              result = T.p(_class="code")[T.a(href=url.URL.fromString(ref))["Starting up"],]
           else:
              if self.hub.paused:
                 p = T.input(type="submit", name="switch", value="Unpause", _class="input.btn")
              else:
                 p = T.input(type="submit", name="switch", value="Pause", _class="input.btn")
              elements = T.span(_class="input.btn")["On"], T.input(type="submit", name="switch", value="Off", _class="input.btn"), p,
              result = T.form(action=url.here.child(SUBMIT), method="post")[ T.table[T.tr[(T.td[el] for el in elements)]]]
        else:
           elements = T.input(type="submit", name="switch", value="On", _class="input.btn"), T.span(_class="input.btn")["Off"], T.span(_class="input.btn")["Pause"],
           result = T.form(action=url.here.child(SUBMIT), method="post")[ T.table[T.tr[(T.td[el] for el in elements)]]]
        return result

    def render_here( self, ctx, data):
        request = inevow.IRequest(ctx)
        ref = request.URLPath()
        new = ref.parentdir()
        return T.a(href=str(new))["here"]

    def render_echo(self,ctx,data):
        obj = EchoElement()
        obj.setFragmentParent(self)
        return ctx.tag[obj]

    def old_makeDocFactory ( self ):
        return loaders.stan(T.html[
                T.head(render=T.directive('liveglue'))[
                    T.title["Live Page"],
                    ],
                T.body[
                     T.div(render=T.directive("echo")),
                     ]
                     ])

    def makeDocFactory ( self ):
        return loaders.stan( T.html[
                  T.head(render=T.directive('liveglue'))[
                        T.title['Hub Page'],
                        ],
                  T.body[
                          T.h2["This is the Hub controller page."],
                          T.div["(The main page is ",T.directive("here"),".)"],
#                          T.p(render = T.directive( "password" )),
                          T.div(render = T.directive( "elements" )),
                          T.div(render = T.directive( "stations" )),
#                          T.div(render=T.directive("echo")),
                   ]
                   ]
            )

    def render_stations( self, ctx, data):
        if not getattr(self,"hub",None): return ""
        if not hasattr(self.hub,"TASKS") or not len(getattr(self.hub,"TASKS")):
          return ""
        if hasattr(self,"hub") and self.hubc.running:
          self.stable = StationsTable(hc=self.hubc,length=self.length)
          self.stable.setFragmentParent(self)
          return ctx.tag[self.stable]
        else:
          return "No hub running"

    def beforeRender(self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn(self, reason):
       if hasattr(self,"stable"):
          self.stable.shutdown()

class StationsTable(athena.LiveElement):

    jsClass = u'EchoThing.EchoWidget'

    def __init__(self,*args,**kwargs):
        self.hubc = kwargs["hc"]
        kwargs.pop("hc")
        self.length = kwargs["length"]
        kwargs.pop("length")
        athena.LiveElement.__init__ (self,*args,**kwargs)
        self.begin()

    def begin(self):
        self.queuing = {}
        self.inLoop = False
        self.hub = self.hubc.hub
        self.queue = {}
        self.index = {}
        self.logpos = {}
        self.loops = {}
        self.inLoop = {}
        self.docFactory = self.makeDocFactory()

    def beforeRender(self, ctx):
        d = self.notifyOnDisconnect()
        d.addErrback(self.disconn)

    def disconn(self, reason):
        reactor.callFromThread(self.clock.stop)

    def makeRecords( self, ctx):
        self.Records = {}
        self.cRecords = {}
        for h in self.hub.TASKS:
           print "Setting up livelog for",h
           self.Records[h] = LiveLog()
           self.Records[h].setHost(h)
           self.Records[h].setFragmentParent(self)
           self.cRecords[h] = self.Records[h]

    def render_table( self, ctx, data):
        self.makeRecords(ctx)
        tables = T.div[[self.cRecords[h] for h in self.hub.TASKS]]
        return tables

    def station( self, h):
        item = u"stn(%d)" % (hash("%r" % (h)))
        p = T.table(style="border:thin solid black;")[
            T.thead[T.tr[T.th(style="border:thin solid black;")[h.ch1.rhost]]],
            T.tbody[T.tr[T.td[T.div(style="height:180px;width:540px;overflow:auto;padding:0px",name=item)]]]
            ]
        return p

    def makeDocFactory( self):
        if hasattr(self,"hub") and self.hub.TASKS:
           ncol = 3
           hs = self.hub.TASKS.keys()
#           s = [self.station(h) for h in self.hub.TASKS]
           n = len(hs)
           r = (n+(ncol-1))/ncol
           print r,n
           t = T.table[
                   self.say(),
                   (T.tr[ \
                        ((j<n and T.td[self.station(hs[j])] or T.td[""]) \
                                for j in range(ncol*k,ncol*(k+1)))
                        ] for k in range(r))
                   ]
        else:
           t = T.div
        return loaders.stan(T.div(render=T.directive('liveElement'))[
            [T.h2["Live Hub"], t]
                ]
                )

    def shutdown(self):
        if hasattr(self,"clock") and self.clock.running:
           reactor.callFromThread(self.clock.stop)

    def say(self):
        for h in self.hub.TASKS:
           self.inLoop[h] = False
           self.queuing[h] = False
           self.index[h] = 0
           self.queue[h] = []
           self.logpos[h] = 0
#           self.loops[h] = task.LoopingCall(self.loop,h)
#           self.loops[h].start(0.05)
        self.queuing0 = False
        self.loop = task.LoopingCall(self.loop)
        self.loop.start(0.05)
        return ""

    def problem(self,reason,h=None):
        if h:
           print "Problem with",h
        logging.err(reason)
        reason.printTraceback(sys.stdout)

    def queueOut(self,h=None):
        if h:
           if not self.queue[h]:
              self.queuing[h] = False
              return False
           else:
              self.queuing[h] = True
              d = defer.Deferred()
              d.addCallback(lambda result: self.callRemote(*(self.queue[h].pop(0))))
              d.addCallback(lambda result: self.queue[h])
              d.addCallback(lambda result: self.queueOut(h))
              d.addErrback(lambda reason: self.problem(reason,h))
              d.callback(True)
        else:
           self.queuing0 = sum([len(self.queue[h]) for h in self.hub.TASKS])>0
           if self.queuing0>0:
              hs = self.hub.TASKS.keys()
              msg = {}
              [msg.setdefault(h,len(self.queue[h])) for h in self.hub.TASKS]
              msgs = max(msg.values())
              send = []
              [send.append(self.queue[h].pop(0)) for l in range(msgs) for h in hs if self.queue[h]]
              e = defer.Deferred()
              for item in send:
                  d = defer.Deferred()
                  d.addCallback(lambda result: self.callRemote(*(send.pop(0))))
                  d.addCallback(lambda result: send)
                  d.addErrback(lambda reason: self.problem(reason))
                  e.chainDeferred(d)
              e.addErrback(lambda reason: self.problem(reason))
              e.callback(True)
              self.queuing0 = False

    def loop(self,h=None):
        if h:
           if self.inLoop[h]: return
           if hasattr(self.page,"callRemote"):
              self.inLoop[h] = True
              adding = 0
              if not self.hubc.hub.TASKS or not hasattr(self.page,"callRemote"):
                 print u"STOPPING LOOPING"
                 reactor.callFromThread(self.loops[h].stop)
                 return
              item = u"stn(%d)" % (hash("%r" % (h)))
              newpos = len(h.ch1.log)
              if newpos > self.logpos[h]:
                 newlog = u"%s" % ( h.ch1.log[self.logpos[h]:newpos])
                 lines = newlog.split("\n")
                 if len(lines) > self.length: lines = lines[-self.length+1:]
                 self.logpos[h] = newpos
                 for message in lines:
                     self.index[h] += 1
                     if self.index[h] > (self.length-0):
                        name = u"%s-clock%d" % (item,self.index[h]-(self.length-0))
                        self.queue[h].append(['delText', item, name])
                     name = u"%s-clock%d" % (item,self.index[h])
                     if self.index[h]==0:
                        self.queue[h].append(['addText', u"div", item, message, name])
                     else:
                        location = u"%s-clock%d" % (item,self.index[h]-1)
                        self.queue[h].append(['stackText', u"div", item, message, name, location])
                     adding += 1
              self.inLoop[h] = False
              if adding and not self.queuing[h]:
                 self.queuing[h] = True
                 self.queueOut(h)
        else:
           if hasattr(self.page,"callRemote"):
              if not self.hubc.hub.TASKS or not hasattr(self.page,"callRemote"):
                 print u"STOPPING LOOPING"
                 reactor.callFromThread(self.loop.stop)
                 return
              adding = 0
              self.queuing0 = sum([len(self.queue[h]) for h in self.hub.TASKS])>0
              for h in self.hub.TASKS:
                 item = u"stn(%d)" % (hash("%r" % (h)))
                 newpos = len(h.ch1.log)
                 if newpos > self.logpos[h]:
                    newlog = u"%s" % ( h.ch1.log[self.logpos[h]:newpos])
                    lines = newlog.split("\n")
                    if len(lines) > self.length: lines = lines[-self.length+1:]
                    self.logpos[h] = newpos
                    for message in lines:
                        self.index[h] += 1
                        if self.index[h] > (self.length-0):
                           name = u"%s-clock%d" % (item,self.index[h]-(self.length-0))
                           self.queue[h].append(['delText', item, name])
                        name = u"%s-clock%d" % (item,self.index[h])
                        if self.index[h]==0:
                           self.queue[h].append(['addText', u"div", item, message, name])
                        else:
                           location = u"%s-clock%d" % (item,self.index[h]-1)
                           self.queue[h].append(['stackText', u"div", item, message, name, location])
                        adding += 1
#              if adding and not self.queuing0:
              if adding:
                 self.queuing0 = True
                 e = defer.Deferred()
                 e.addCallback(lambda x: self.queueOut())
                 e.addErrback(lambda reason: self.problem(reason))
                 e.callback(True)



class HubControl:
    def __init__(self):
        self.running = 0
        self.start = 0
        self.stop = 0
        self.pause = 0
        self.getPassword()

    def getPassword( self):
        if not hasattr(self,"password"):
           self.password = getpass.getpass("%s's password:" % (user))

    def hubstart( self):
        if self.running:
           pass
        else:
          self.hub = TwistedHub(tasks=[],quiet=quiet,verb=verb,clean=0,wait=1-nowait,norun=0,\
             stayalive=1,user=user,password=self.password,listen=port,sshport=sshport,modules=modules,fmodules=fmodules,\
             machines=machines,nice=nicelevel,sleeptime=stime,namespace=globals(),auth=auth,standalone=False)
          self.running = 1

    def hubstop( self):
        if self.running:
           if self.hub.paused: self.hub.paused = 0
           self.hub.finish = 1
           self.running = 0
        else:
           pass

    def hubpause( self):
        if self.running:
           if self.hub.paused: self.hub.paused = 0
           else: self.hub.paused = 1
        else:
           pass

if __name__ == "__main__":

   HubController = HubControl()
   site = appserver.NevowSite(MainPage(hc=HubController))

   application = service.Application("Echo Demo")
   reactor.listenTCP(7070,site)
   reactor.run()


