
from sys import *
from Multipack import *
from numpy.oldnumeric.mlab import *
import gzip
from basics import *
from numpy.core.numerictypes import *
import numpy
from scipy.special import *
from glob import glob
from pgplot import *
from FITS import *
import pickle
import cPickle
import os
from sets import *
from scipy.interpolate import *

#udir = '/Volumes/Data/kelson/GalMod/Maraston08'
udir = os.environ["PYTHONDATA"]+"/modeling/Maraston08"

def load1prop(prop,hb="red"):
    files = [
       udir+"/hb"+hb+"_kroupa.tab",
       udir+"/hb"+hb+"_kroupa2.tab",
       udir+"/stellarmass.kroupa",
#       udir+"/colorsSDSS_pickles_sspM08",
#       udir+"/colors_pickles_sspM08",
       ]
    for file in files:
        data = readtable(file,header="Z/H")
        if prop in data: break
    if "t(Gyr)" in data: tl = "t(Gyr)"
    elif "t/Gyr" in data: tl = "t/Gyr"
    lt = log10(data[tl])+9.
    lz = data["Z/H"]
    prop = data[prop]
    return lt,lz,prop

class SSPProperty:
   def __init__(self,prop,kx,ky,Z=None,l=0,verb=0,hb="red"):
      self.prop = prop
      self.Z = Z
      self.load(l,hb=hb)
      self.fit(kx,ky,verb=verb)
   def load(self,l,hb="red"):
      a,z,o = load1prop(self.prop,hb=hb)
      if equal(z,self.Z).any():
         k = equal(z,self.Z)
      else:
         k = between(z,-2.0,0.4)
      a,z,o = compress(k,[a,z,o],1)
      sz = sorted(list(Set(z)))
      sa = sorted(list(Set(a)))
      s = len(sz),len(sa)
      a = reshape(a,s)
      z = reshape(z,s)
      o = reshape(o,s)
      self.z = z
      self.a = a
      self.o = o
      if l: self.o = getlog10(self.o)
   def fit(self,kx=5,ky=5,verb=0):
      if self.z.shape[0] > 1:
         tx = self.a[0]
         ty = self.z[::,0]
         while len(ty) < 2*ky+2: ty = concatenate([ty[:1],ty,ty[-1:]])
         if verb:
            print tx
            print ty
         self.tck = bisplrep(self.a.ravel(),self.z.ravel(),self.o.ravel(),kx=kx,ky=ky,s=0,\
           task=-1,tx=tx,ty=ty,nxest=2*len(tx),nyest=2*len(ty),quiet=0,full_output=1)
      else:
         tx = self.a[0]
         if verb:
            print tx
#         self.tck = splrep(self.a.ravel(),self.o.ravel(),k=kx,s=0,t=tx,task=-1,quiet=0,full_output=1)
         self.tck = splrep(self.a.ravel(),self.o.ravel(),k=kx,s=0,task=0,quiet=0,full_output=1)

   def interp(self,lage,feh):
      lage = asarray(lage)
      feh = asarray(feh)
      if self.z.shape[0] > 1:
         if self.Z != None:
           if type(lage) == numpy.ndarray:
             feh = self.Z*ones(lage.shape)
           elif type(lage) == ListType:
             feh = len(lage)*[self.Z]
           else:
             feh = self.Z
         if type(lage) == numpy.ndarray and type(feh) == numpy.ndarray and lage.shape == feh.shape and lage.shape:
           return reshape(bisplev2(lage.ravel(),feh.ravel(),self.tck[0]), lage.shape)
         elif type(lage) == numpy.ndarray and type(feh) == numpy.ndarray and lage.shape > feh.shape and lage.shape:
           return bisplev2(lage,feh*ones(lage.shape),self.tck[0])
         elif type(lage) == numpy.ndarray and type(feh) == numpy.ndarray and lage.shape < feh.shape and feh.shape:
           return bisplev2(lage*ones(feh.shape),feh,self.tck[0])
         else:
           val = bisplev([lage],[feh],self.tck[0])
           return val
      else:
         if type(lage) == numpy.ndarray and lage.shape:
           return reshape(splev(lage.ravel(),self.tck[0]),lage.shape)
         else:
           val = splev([lage],self.tck[0])
           return val

   def tau(self,lage,lend,tau,Z,ninc=16,pmag=0,plog=0,pinv=0,emin=0,mage=0):
      if type(lage) == numpy.ndarray:
         v = array([self.tau1(lage[j],lend[j],tau[j],Z[j],ninc=ninc,pmag=pmag,plog=plog,pinv=pinv,emin=emin,mage=mage) for j in range(len(lage))])
      else:
         v = self.tau1(lage,lend,tau,Z,ninc=ninc,pmag=pmag,plog=plog,emin=emin,mage=mage)
      return v
   def tau1(self,lage,lend,ltau,Z,ninc=16,pmag=0,plog=0,pinv=0,mage=0,emin=5):
      start = pow(10.0,lage)
      finish = max([ exp10(lend),exp10(emin)])
      tau = exp10(ltau)
      if type(Z) is StringType: Zu = lambda t: eval(Z)
      elif type(Z) is FloatType: Zu = lambda t: Z
      else: Zu = lambda t: Z
      if finish<start and ltau:
#         if (start-finish) > 5*tau: finish = start - 5*tau
#         dt = min([tau/(ninc/2),(start-finish)/ninc])
#         ninc2 = int((start-finish)/dt+1)
         dt = (log10(start)-log10(finish))/ninc
         times = exp10(arange(log10(finish),log10(start)+dt/100,dt,float32))
         ltimes = log(times)
         dt = (times[2:]-times[:-2])/2.
         dt = concatenate([dt[:1],dt,dt[-1:]])
         dlt = (ltimes[2:]-ltimes[:-2])/2.
         dlt = concatenate([dlt[:1],dlt,dlt[-1:]])
         dt = times*dlt
         SF = lambda t: exp(-(start-t)/tau)
         SFs = array([SF(moment) for moment in times])
         SFtot = sum(dt*SFs)
         SFs = dt*SFs/SFtot
         Ps = array([self.interp(log10(moment),Zu(moment)) for moment in times])
         if mage:
            if pmag:
              vals = exp10(sum(SFs*exp10(-0.4*Ps)*log10(times)) / sum(SFs*exp10(-0.4*Ps)))
            elif plog:
              vals = exp10(sum(SFs*exp10(Ps)*log10(times)) / sum(SFs*exp10(Ps)))
            else:
              vals = exp10(sum(SFs*log10(times))/sum(SFs))
         elif pmag:
            vals =-2.5*log10(sum(SFs*exp10(-0.4*Ps)))
         elif plog>0:
            vals = log10(sum(SFs*exp10(Ps)))
         elif plog<0:
            vals = -log10(sum(SFs*exp10(-Ps)))
         elif pinv:
            vals = 1/sum(SFs/Ps)
         else:
            vals = sum(SFs*Ps)
      elif lage:
         vals = self.interp(lage,Zu(start))
      else:
         return -99.999
      return vals

class SSPDerivedProperty(SSPProperty):
   def __init__(self,prop1,prop2,func,kx,ky,Z=None,l=0,verb=0,hb="red"):
      self.prop1 = prop1
      self.prop2 = prop2
      self.hb = hb
      self.func = func
      self.prop = "%s(%s,%s)" % (self.func,self.prop1,self.prop2)
      self.Z = Z
      self.kx = kx
      self.ky = ky
      self.load(l)
      self.fit(kx,ky,verb=verb)
   def load(self,l):
      p1 = SSPProperty(self.prop1,self.kx,self.ky,Z=self.Z,l=l,verb=0,hb=self.hb)
      p2 = SSPProperty(self.prop2,self.kx,self.ky,Z=self.Z,l=l,verb=0,hb=self.hb)
      if self.Z == None:
         a = p1.a
         z = p1.z
         o1 = p1.o
         o2 = p2.o
      else:
         o1 = p1.interp(p1.a.ravel(),self.Z*ones(p1.a.ravel().shape))
         o2 = p2.interp(p1.a.ravel(),self.Z*ones(p1.a.ravel().shape))
         a = p1.a.ravel()
         z = self.Z*ones(a.shape)
         o1 = reshape(o1,p1.a.shape)
         o2 = reshape(o1,p1.a.shape)
         a = reshape(a,p1.a.shape)
         z = reshape(z,p1.a.shape)
      o = self.func(o1,o2)
      self.z = z
      self.a = a
      self.o = o
      if l: self.o = getlog10(self.o)






